/*
 * main.c:
 * Simple conversion program for coordinates.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] =
    "$Id: libcoords.c,v 1.6 2001/12/05 02:30:49 root Exp $";

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "libcoords.h"
#include "coords.h"

IOSGRID wgs84_to_iosgrid(WGS84 w)
{
  IOSGRID g = { 0 };
  char zone[3];
  double east, north;


  //WGS84 -> Airy
  wgs84_to_geog(&w.lat, &w.lon, DATUM_IRISH);



  /* Obtain OS grid reference. */
  if (!geog_to_irish(w.lat, w.lon, zone, &east, &north, DATUM_WGS84))
    return g;



  g.zone = zone[0];
  g.e = (int) (east + .5);
  g.n = (int) (north + .5);

  return g;
}

BOSGRID wgs84_to_bosgrid(WGS84 w)
{
  BOSGRID g = { 0 };
  char zone[3];

  double east, north;


  //WGS84 -> Airy
  wgs84_to_geog(&w.lat, &w.lon, DATUM_OSGB);

  /* Obtain OS grid reference. */
  if (!geog_to_osgb(w.lat, w.lon, zone, &east, &north, DATUM_WGS84))
    return g;


  g.big = zone[0];
  g.med = zone[1];
  g.e = (int) (east + .5);
  g.n = (int) (north + .5);

  return g;
}

WGS84 bosgrid_to_wgs84(BOSGRID g)
{
  WGS84 w = { 0 };
  char zone[3];

  double east, north;

  zone[0] = g.big;
  zone[1] = g.med;
  zone[2] = 0;
  east = g.e;
  north = g.n;

  /* OS -> AIRY */
  if (!osgb_to_geog(zone, east, north, &w.lat, &w.lon, DATUM_WGS84))
    return w;

  //Airy -> WGS84
  geog_to_wgs84(&w.lat, &w.lon, DATUM_OSGB);

  return w;
}

WGS84 iosgrid_to_wgs84(IOSGRID g)
{
  WGS84 w = { 0 };
  char zone[3];
  double east, north;

  zone[0] = g.zone;
  zone[1] = 0;
  east = g.e;
  north = g.n;

  /* IRISH->AIRY */
  if (!irish_to_geog(zone, east, north, &w.lat, &w.lon, DATUM_WGS84))
    return w;


  //Airy -> WGS84
  geog_to_wgs84(&w.lat, &w.lon, DATUM_IRISH);

  return w;
}
