#include "project.h"
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <termio.h>
#include <math.h>
#include <time.h>

#define K1 0.000000059604644775390625L
#define K2 0.00390625L
#define R2D (180.0/3.141592654)
#define D2R (3.141592654/180.0)

int state;
unsigned char mdata[10];
unsigned char buf[1024];
int dwc, dwp;
FILE *where;




static void fix_flip_2(unsigned char *addr, int n)
{
  unsigned char tmp;

  for (; n > 0; --n) {
    tmp = *addr;
    *addr = *(addr + 1);
    addr++;
    *addr = tmp;
    addr++;
  }


}

static void fix_flip_4(unsigned char *addr, int n)
{
  unsigned char tmp;

  for (; n > 0; --n) {
    tmp = *addr;
    *addr = *(addr + 3);
    *(addr + 3) = tmp;
    addr++;
    tmp = *addr;
    *addr = *(addr + 1);
    addr++;
    *addr = tmp;
    addr += 2;
  }


}





short int put_UI16(unsigned char *ptr, unsigned short int val)
{
  bcopy((unsigned char *) &val, ptr, sizeof(val));

#ifdef _BIG_ENDIAN
  fix_flip_2((unsigned char *) ptr, 1);
#endif
}

short int put_I16(unsigned char *ptr, short int val)
{
  bcopy((unsigned char *) &val, ptr, sizeof(val));

#ifdef _BIG_ENDIAN
  fix_flip_2((unsigned char *) ptr, 1);
#endif
}

short int get_I16(char *ptr)
{
  short int ret;

  bcopy(ptr, (unsigned char *) &ret, sizeof(ret));

#ifdef _BIG_ENDIAN
  fix_flip_2((unsigned char *) &ret, 1);
#endif


  return (ret);
}

unsigned short int get_UI16(char *ptr)
{
  short int ret;

  bcopy(ptr, (unsigned char *) &ret, sizeof(ret));

#ifdef _BIG_ENDIAN
  fix_flip_2((unsigned char *) &ret, 1);
#endif


  return (ret);
}

long int get_I32(char *ptr)
{
  int ret;
  unsigned char *p = (unsigned char *) &ret;

  bcopy(ptr, (unsigned char *) &ret, sizeof(ret));

#ifdef _BIG_ENDIAN
  fix_flip_4((unsigned char *) &ret, 1);
#endif

  return (ret);
}

long unsigned int get_UI32(char *ptr)
{
  unsigned int ret;
  unsigned char *p = (unsigned char *) &ret;

  bcopy(ptr, (unsigned char *) &ret, sizeof(ret));

#ifdef _BIG_ENDIAN
  fix_flip_4((unsigned char *) &ret, 1);
#endif

  return (ret);
}


float get_F32(unsigned char *ptr)
{
  double mant, exp, sign, ret;
  int tmp;
  long unsigned int lv;
  unsigned char *pt2 = ptr;

  int j;

  tmp = 0;
  for (j = 0; j < 4; ++j)
    tmp |= *(pt2++);

  if (tmp == 0)
    return (0.0);

  pt2--;

  if (*pt2 & 0x80)
    sign = -1.0;
  else
    sign = 1;

  lv = *(pt2--);
  lv <<= 8;
  lv |= *(pt2--);
  lv <<= 8;
  lv |= *(pt2--);
  lv |= 0x800000;

  mant = K1 * (double) lv;
  exp = (double) ((*ptr) - 128);

  ret = mant * pow(2.0, exp) * sign;
  return ((float) ret);

}

double get_F48(unsigned char *ptr)
{
  double mant, exp, sign, ret;
  int tmp;
  long unsigned int lv;
  unsigned char *pt2 = ptr;

  int j;

  tmp = 0;
  for (j = 0; j < 6; ++j)
    tmp |= *(pt2++);

  if (tmp == 0)
    return (0.0);

  pt2--;

  if (*pt2 & 0x80)
    sign = -1.0;
  else
    sign = 1;

  pt2 = ptr + 1;

  mant = (double) (*pt2++);
  mant *= K2;
  mant += (double) (*pt2++);
  mant *= K2;
  mant += (double) (*pt2++);
  mant *= K2;
  mant += (double) (*pt2++);
  mant *= K2;
  mant += (double) ((*pt2++) & 0x7f);
  mant *= K2;
  mant += 0.5;

  exp = (double) ((*ptr) - 128);

  ret = mant * pow(2.0, exp) * sign;
  return (ret);

}


typedef union {
  unsigned char bd[8];
  int iv;
  long int lv;
  float fv;
  double dv;
} Value;

put_F48(unsigned char *ptr, double val)
{
  int i;
  Value valu;

  valu.dv = val;

  if (val == 0.0)
    bzero(ptr, 8);
  else {

#ifdef _BIG_ENDIAN
    fix_flip_8((unsigned char *) &val, 1);
#endif

    ptr[5] = valu.bd[7] & 0x80;
    ptr[0] = (valu.bd[7] & 0x40) << 1;
    ptr[0] |= (valu.bd[7] & 0x07) << 4;
    ptr[0] |= (valu.bd[6] & 0xf0) >> 4;

    ptr[0] = (ptr[0] - 0x7e) ^ 0x80;

    ptr[5] |= (valu.bd[6] & 0x0f) << 3;
    ptr[5] |= (valu.bd[5] & 0xe0) >> 5;

    ptr[4] = (valu.bd[5] & 0x1f) << 3;
    ptr[4] |= (valu.bd[4] & 0xe0) >> 5;

    ptr[3] = (valu.bd[4] & 0x1f) << 3;
    ptr[3] |= (valu.bd[3] & 0xe0) >> 5;

    ptr[2] = (valu.bd[3] & 0x1f) << 3;
    ptr[2] |= (valu.bd[2] & 0xe0) >> 5;

    ptr[1] = (valu.bd[2] & 0x1f) << 3;
    ptr[1] |= (valu.bd[1] & 0xe0) >> 5;
  }
}

put_F32(unsigned char *ptr, float val)
{
  Value valu;

  valu.fv = val;

  if (valu.fv == 0.0)
    bzero(ptr, 4);
  else {

#ifdef _BIG_ENDIAN
    fix_flip_4(ptr, 1);
#endif

    ptr[3] = valu.bd[3] & 0x80;

    ptr[0] = (valu.bd[3] & 0x7f) << 1;
    if (valu.bd[2] & 0x80)
      ptr[0]++;

    ptr[3] |= (valu.bd[2] & 0x7f);
    ptr[2] = valu.bd[1];
    ptr[1] = valu.bd[0];

    ptr[0] = (ptr[0] - 0x7e) ^ 0x80;
  }

}

int checksum(unsigned char *ucdata, int len)
{
  unsigned short int *data = (unsigned short int *) ucdata;
  int sum = 0;
  unsigned short int moo;
  short int *cow = (short int *) &moo;
  len >>= 1;

  while (len) {
    sum += *(data++);
    len--;
    sum %= 65536;
  }

  moo = sum;

  if (*cow != -32768)
    *cow = -*cow;
  return (*cow);
}

char str_chan_buf[1024];
char *str_chan(unsigned char *pkt)
{
  int stat;
  int cn;
  int i;

  char buf[1024];


  strcpy(str_chan_buf, "");

  stat = get_I16(pkt);
  cn = get_I16(pkt + 2);

#ifdef _BIG_ENDIAN
  fix_flip_2(&stat, 1);
  fix_flip_2(&cn, 1);
#endif

  sprintf(str_chan_buf, "Satalite %d ", stat & 63);

  stat >>= 9;
  i = stat & 7;

  if (i == 0) {
    strcat(str_chan_buf, "Idle ");
  } else if (i == 1) {
    strcat(str_chan_buf, "C/A Search ");
  } else if (i == 5) {
    strcat(str_chan_buf, "C/A Tracking ");
  } else {
    strcat(str_chan_buf, "Unknown status ");
  }

  stat >>= 3;
  stat &= 7;
  sprintf(buf, "Ch %d ", stat);
  strcat(str_chan_buf, buf);


  cn >>= 8;
  cn &= 63;

  sprintf(buf, "C/N %d dBHz", cn);
  strcat(str_chan_buf, buf);


  return (str_chan_buf);
}

char *str_soh(unsigned char *pkt)
{
  int i = get_I16(pkt);


  strcpy(str_chan_buf, "");

  if (i & 1)
    strcat(str_chan_buf, "Power cycling, ");
  if (i & 4)
    strcat(str_chan_buf, "SRAM b fail, ");
  if (i & 8)
    strcat(str_chan_buf, "RTC b fail, ");

  return (str_chan_buf);
}

char *str_stat(unsigned char *pkt)
{
  int i = get_I16(pkt);
  int j;
  char buf[100];

  strcpy(str_chan_buf, "");

  if (i & 1)
    strcat(str_chan_buf, "Alt acpt, ");
  if (i & 2)
    strcat(str_chan_buf, "Alt used, ");
  if (i & 4)
    strcat(str_chan_buf, "Alt hold, ");
  if (i & 8)
    strcat(str_chan_buf, "Nav, ");
  else
    strcat(str_chan_buf, "Aqusn, ");

  j = i;
  j >>= 4;
  j &= 7;


  if (i & 128)
    strcat(str_chan_buf, "cold start. ");
  else
    strcat(str_chan_buf, "warm start. ");

  sprintf(buf, "\nUsed %d sats, ", j);
  strcat(str_chan_buf, buf);

  i >>= 8;
  j = i & 15;

  switch (j) {
  case 1:
    strcat(str_chan_buf, "err<7.5,");
    break;
  case 2:
    strcat(str_chan_buf, "7.5<err<12.5,");
    break;
  case 3:
    strcat(str_chan_buf, "12.5<err<25,");
    break;
  case 4:
    strcat(str_chan_buf, "25<err<50,");
    break;
  case 5:
    strcat(str_chan_buf, "50<err<250,");
    break;
  case 6:
    strcat(str_chan_buf, "250<err<500,");
    break;
  case 7:
    strcat(str_chan_buf, "500<err,");
    break;
  case 9:
    strcat(str_chan_buf, "converg error,");
    break;
  default:
    strcat(str_chan_buf, "error is unknown,");


  }
  j = i >> 4;
  j &= 7;
  sprintf(buf, "\nUtil ch %d , ", j);
  strcat(str_chan_buf, buf);

  if (i & 128)
    strcat(str_chan_buf, "DGPS.");
  else
    strcat(str_chan_buf, "no DGPS.");

  return (str_chan_buf);

}



int dtr = 0, dtre = 0;

process_packet(int fd, unsigned short mid, unsigned char *pkt, int len,
               int type)
{
  int i;
  switch (type) {
  case 0:
    printf("Received packet %d len %d\n", mid, len);
    break;
  case 0x0800:
    printf("GPS +ACK for packet %d\n", mid);
    break;
  case 0x0040:
    printf("GPS -NACK for packet %d\n", mid);
    break;
  default:
    printf("Packet %d len %d unknown type %04x\n", mid, len, type);
    break;
  }

/*
for (i=0;i<len;++i) {
printf("%02x",pkt[i]);
}
printf("\n");
*/




  switch (mid) {
  case 101:
    printf("BIT results:\n");
    printf("summary %04x\n", get_UI16(pkt));
    printf("low order ram    %04x\n", get_UI16(pkt + 2));
    printf("high order ram   %04x\n", get_UI16(pkt + 4));
    printf("preprocessor     %04x\n", get_UI16(pkt + 6));
    printf("preprocessor ram %04x\n", get_UI16(pkt + 8));
    printf("multifunction if %04x\n", get_UI16(pkt + 10));
    printf("Code/Carrier VCO %04x\n", get_UI16(pkt + 12));
    printf("Code Gen./Signal %04x\n", get_UI16(pkt + 14));
    printf("Serial error count   %d\n", get_I16(pkt + 16));
    printf("Serial chksum errors %d\n", get_I16(pkt + 18));
    printf("EEPROM prog errors   %d\n", get_I16(pkt + 20));
    printf("EEPROM chksum errors %d\n", get_I16(pkt + 22));
    printf("AUX XMIT error count %d\n", get_I16(pkt + 24));
    printf("software config  %04x\n", get_UI16(pkt + 26));
    printf("software version %.3f\n", get_F32(pkt + 28));
    break;
  case 102:
    printf("Set time\t%d\n", get_UI32(pkt));
    printf("GPS tow\t%f\n", get_F48(pkt + 4));
    printf("GPS wk\t%d\n", get_I16(pkt + 10));
    {
      int n = get_I16(pkt + 12);
      unsigned char *ptr = pkt + 14;

      printf("Number of sats\t%d\n", n);

      while (n--) {

        printf("Sat %02d El %02.1f Az %03.1f\n",
               get_I16(ptr),
               get_F32(ptr + 2) * R2D, get_F32(ptr + 6) * R2D);

        ptr += 10;

      }
    }
    break;
  case 103:

    printf("GPS tow\t%f\n", get_F48(pkt));
    printf("GPS wk\t%d\n", get_I16(pkt + 6));
    printf("UTC val\t%d\n", get_I16(pkt + 8));
    printf("UTC tod\t%f\n", get_F48(pkt + 10));
    printf("UTC day\t%d\n", get_I16(pkt + 16));
    printf("UTC mon\t%d\n", get_I16(pkt + 18));
    printf("UTC yr\t%d\n", get_I16(pkt + 20));
    printf("UTC set\t%f\n", 0.02 * (float) get_I32(pkt + 22));
    printf("X\t%f\tY\t%f\tZ\t%f\n",
           get_F48(pkt + 26), get_F48(pkt + 32), get_F48(pkt + 38));
    printf("VE\t%f\tVN\t%f\tVU\t%f\n",
           get_F32(pkt + 44), get_F32(pkt + 48), get_F32(pkt + 52));

    update_speeds(get_F32(pkt + 44),
                  get_F32(pkt + 48), get_F32(pkt + 52), get_F32(pkt + 68));
    printf("Lat\t%f\tLon\t%f\n",
           R2D * get_F48(pkt + 56), R2D * get_F48(pkt + 62));
    {
      time_t tme;
#ifdef PLAYBACK
      float lat, lon;
      fscanf(where, "%d %f %f\n", &tme, &lat, &lon);
      printf("%d %f %f\n", tme, lat, lon);
      pos.lat = lat;
      pos.lon = lon;

#else
      time(&tme);
      fprintf(where, "%d %30.20f %30.20f\n", tme,
              R2D * get_F48(pkt + 56), R2D * get_F48(pkt + 62));
      fflush(where);

      pos.lat = R2D * get_F48(pkt + 56);
      pos.lon = R2D * get_F48(pkt + 62);
#endif
    }

    new_pos();

    printf("Alt\t%f\n", get_F32(pkt + 68));
    printf("HDOP\t%d\tVDOP\t%d\tPDOP\t%d\n",
           get_I16(pkt + 72), get_I16(pkt + 74), get_I16(pkt + 76));
    printf("TDOP\t%d\tGDOP\t%d\n", get_I16(pkt + 78), get_I16(pkt + 80));
    printf("C1 %s\n", str_chan(pkt + 82));
    printf("C2 %s\n", str_chan(pkt + 86));
    printf("C3 %s\n", str_chan(pkt + 90));
    printf("C4 %s\n", str_chan(pkt + 94));
    printf("C5 %s\n", str_chan(pkt + 98));
    printf("EHPE\t%d\tEVPE\t%d\n", get_I16(pkt + 102), get_I16(pkt + 104));
    printf("ETE\t%d\tEHVE\t%d\n", get_I16(pkt + 106), get_I16(pkt + 108));
    printf("SOH %s\n", str_soh(pkt + 112));
    printf("%s\n", str_stat(pkt + 114));
    break;
  default:
  }

}


send_message(int fd, int mid, int ack, unsigned char *pkt, int len)
{
  unsigned char *buf, *ptr;

  ptr = buf = (unsigned char *) malloc(len + 12);

  len >>= 1;

  *(ptr++) = 0xff;
  *(ptr++) = 0x81;
  put_I16(ptr, mid);
  ptr += 2;
  put_I16(ptr, len);
  ptr += 2;
  *(ptr++) = 0x00;
  if (ack)
    *(ptr++) = 0x10;
  else
    *(ptr++) = 0x0;
  put_I16(ptr, checksum(buf, 8));
  ptr += 2;
  len <<= 1;
  bcopy(pkt, ptr, len);
  ptr += len;
  put_I16(ptr, checksum(pkt, len));

//write(fd,buf,len+12);
  free(buf);
}


/*select_satellites(int fd)
{
unsigned char pkt[6];
put_I16(pkt,1);
put_UI16(pkt+2,S16 | S02 | S04 | S07 | S04 | S05);
put_UI16(pkt+4,S24);
send_message(fd,208,1,pkt,6);
}
*/

request_bit(int fd)
{
  unsigned char pkt[4];
  bzero(pkt, 4);
  send_message(fd, 202, 1, pkt, 4);
}

init_gps(int fd, double lat, double lon, float alt)
{
  unsigned char pkt[60];

  time_t secs;
  struct tm *utc;

  lat *= D2R;
  lon *= D2R;

  secs = time(NULL);
  utc = gmtime(&secs);

  bzero(pkt, 60);

  put_I16(pkt, 0x00fe);         /* Reset 3 sat/auto reliable lat/utc/vel */
  put_F48(pkt + 20, lat);
  printf("%f=?=%f\n", get_F48(pkt + 20), lat);
  put_F48(pkt + 26, lon);
  printf("%f=?=%f\n", get_F48(pkt + 26), lon);
  put_F32(pkt + 32, alt);
  printf("%f=?=%f\n", get_F32(pkt + 32), alt);
  put_F32(pkt + 36, 0.0);
  put_F32(pkt + 40, 0.0);
  put_F32(pkt + 44, 0.0);
  printf("%d %d %d %d %d %d", utc->tm_hour, utc->tm_min, utc->tm_sec,
         utc->tm_mday, utc->tm_mon, utc->tm_year);
  put_I16(pkt + 48, utc->tm_hour);
  put_I16(pkt + 50, utc->tm_min);
  put_I16(pkt + 52, utc->tm_sec);
  put_I16(pkt + 54, utc->tm_mday);
  put_I16(pkt + 56, utc->tm_mon);
  put_I16(pkt + 58, utc->tm_year);

  send_message(fd, 201, 1, pkt, 60);

}


process_char(int fd, unsigned char c)
{
  unsigned short mid;

  switch (state) {
  case 0:
    mdata[0] = mdata[1];
    mdata[1] = c;

    if ((mdata[0] == 0xff) && (mdata[1] == 0x81))
      state += 2;
    break;
  case 2:
  case 3:
  case 4:
  case 5:
  case 6:
  case 7:
  case 8:
    mdata[state++] = c;
    break;
  case 9:
    mdata[state++] = c;
    dwc = get_I16(mdata + 4);
    dwc <<= 1;

// printf("Message header reveiced %d data words\n",dwc);

    dwc += 2;
    dwp = 0;

    break;
  case 10:
    dwc--;
    buf[dwp++] = c;
    if (!dwc)
      state++;
    break;
  case 11:
    mid = get_I16(mdata + 2);

    process_packet(fd, mid, buf, dwp, get_I16(mdata + 6));

    state = 0;
    break;
  default:
    state = 0;
  }


}



static int gps_fd, log_fd;

int gps_init(char *dev)
{
  unsigned char c;
  int fd;
  struct termios tios;
  int speed;


  log_fd = open("test.log", O_RDONLY);
  lseek(log_fd, 100000L, SEEK_SET);

#ifdef PLAYBACK
  where = fopen("play.ben", "r");
#else
  where = fopen("where.ben", "a");
#endif

//  fd = open ("/dev/ttyS0", O_RDWR);
  fd = open(dev, O_RDONLY);

  if (fd <= 0) {
    fprintf(stderr, "Failed to open GPS device %s\n", dev);
    exit(1);
  }

  tcgetattr(fd, &tios);

  tios.c_cflag &= ~(CSIZE | CSTOPB | PARENB | CLOCAL | CRTSCTS);
  tios.c_cflag |= CLOCAL;
  tios.c_cflag |= CS8;

  tios.c_cflag |= CREAD | HUPCL;
  tios.c_iflag = IGNBRK | IGNPAR;

  tios.c_oflag = 0;
  tios.c_lflag = 0;
  tios.c_cc[VMIN] = 1;
  tios.c_cc[VTIME] = 0;

//Fixme: maybe a numeric to Bxxxx speed converter switch stm here...

  speed = B4800;

  if (speed) {
    cfsetospeed(&tios, speed);
    cfsetispeed(&tios, speed);
  }


  if (tcsetattr(fd, TCSAFLUSH, &tios) < 0)
    perror("Error setting communication parameters");

  gps_fd = fd;

  return fd;
}

void gps_dispatch(void)
{
  unsigned char c;

  read(gps_fd, &c, 1);
  process_char(gps_fd, c);

/*
read(log_fd,&c,1);
read(log_fd,&c,1);
process_char(gps_fd,c);
read(log_fd,&c,1);
process_char(gps_fd,c);
read(log_fd,&c,1);
process_char(gps_fd,c);
read(log_fd,&c,1);
process_char(gps_fd,c);
*/
}
