#include "project.h"


static letter_to_x(char c)
{
  switch (c | 32) {
  case 'a':
  case 'f':
  case 'l':
  case 'q':
  case 'v':
    return 0;
  case 'b':
  case 'g':
  case 'm':
  case 'r':
  case 'w':
    return 1;
  case 'c':
  case 'h':
  case 'n':
  case 's':
  case 'x':
    return 2;
  case 'd':
  case 'j':
  case 'o':
  case 't':
  case 'y':
    return 3;
  }
}


static letter_to_y(char c)
{
  switch (c | 32) {
  case 'a':
  case 'b':
  case 'c':
  case 'd':
    return 4;
  case 'f':
  case 'g':
  case 'h':
  case 'j':
    return 3;
  case 'l':
  case 'm':
  case 'n':
  case 'o':
    return 2;
  case 'q':
  case 'r':
  case 's':
  case 't':
    return 1;
  case 'v':
  case 'w':
  case 'x':
  case 'y':
    return 0;
  }
}

static char xy_to_letter(int x, int y)
{
  char tbl[5][4] = { {'V', 'W', 'X', 'Y'},
  {'Q', 'R', 'S', 'T'},
  {'L', 'M', 'N', 'O'},
  {'F', 'G', 'H', 'J'},
  {'A', 'B', 'C', 'D'}
  };

  if (x < 0)
    return ' ';
  if (y < 0)
    return ' ';
  if (x > 3)
    return ' ';
  if (y > 4)
    return ' ';

  return tbl[y][x];
}

print_iosxy(IOSXY o)
{
  printf("%ld %ld\n", o.x, o.y);
}

print_iosgrid(IOSGRID g)
{
  printf("%c%d%d\n", g.zone, g.e, g.n);
}


sprint_iosgrid(char *buf, IOSGRID g)
{
  sprintf(buf, "%c %05d %05d", g.zone, g.e, g.n);
}

print_iosmap(IOSMAP m)
{
  printf("%d.%d %f %f\n", m.lx, m.ty, m.x, m.y);
}


IOSMAP iosxy_to_iosmap(IOSXY x)
{
  IOSMAP m;


  x.y -= 1122;

  m.lx = x.x - 6491;
  m.ty = x.y - 1955;

  m.lx = m.lx / 7800;
  m.ty = m.ty / 5780;

  m.lx *= 7800;
  m.ty *= 5780;


  m.lx += 6491;
  m.ty += 1955;


  x.x -= m.lx;
  x.y -= m.ty;

  m.ty += 5780;

  m.x = ((float) x.x) / 7800.f;
  m.y = ((float) x.y) / 5780.f;

  return m;
}

IOSXY iosmap_to_iosxy(IOSMAP m)
{
  IOSXY x;

  m.ty -= 5780;

  x.x = m.lx;
  x.y = m.ty;


  x.x += (int) ((7800.f * m.x) + .5f);
  x.y += (int) ((5780.f * m.y) + .5f);

  x.y += 1122;

  return x;
}


IOSXY iosgrid_to_iosxy(IOSGRID g)
{
  IOSXY o;

  o.x = g.e;
  o.y = g.n;

  o.x += 100000 * letter_to_x(g.zone);
  o.y += 100000 * letter_to_y(g.zone);


  return o;
}


IOSGRID iosxy_to_iosgrid(IOSXY o)
{
  int x, y;
  IOSGRID g;

  x = o.x / 100000;
  y = o.y / 100000;

  g.zone = xy_to_letter(x, y);

  o.x -= x * 100000;
  o.y -= y * 100000;

  g.e = o.x;
  g.n = o.y;

  return g;
}

IOSXY iosxy_neighbour(IOSXY o, int n, int e)
{
  o.x += 7800L * (long) n;
  o.y += 5780L * (long) e;
  return o;
}

char *iosmap_to_filename(IOSMAP m)
{
  char buf[1024];
  sprintf(buf, "%s/%06d/%06d.%06d.jpg", getenv("IOSPATH"), m.lx, m.lx,
          m.ty);

  return strdup(buf);

}

int ios_same_map(IOSMAP m1, IOSMAP m2)
{
  if (m1.ty != m2.ty)
    return 0;
  if (m1.lx != m2.lx)
    return 0;
  return 1;
}
