#include "project.h"

Display *dpy;
Window win;
GC gc;

Frame frame;
Canvas canvas;
Panel panel;
Panel_item scale, placeitem, head, mps, mph, alt, gridref, dist, eta,
    destref;
Panel_item motion_speed;



WGS84 pos;

//WGS84 pos = {57.4396183333,-6.6850316666}; // Skye
//WGS84 pos = { 55.0119683333, -8.0108616666 }; // Ireland

Places place;

#ifdef FISHSOUP
BOSMAP fishsoup = { 0 };
#endif

static void canvas_ev(Canvas ec, Event * e);


static void do_motion(int e, int n)
{
  char buf[1024];
  switch (place) {

  case IRELAND:
    {
      IOSGRID ng, g = wgs84_to_iosgrid(pos);
      IOSXY nxy, xy = iosgrid_to_iosxy(g);
      nxy = iosxy_neighbour(xy, e, n);
      ng = iosxy_to_iosgrid(nxy);
      pos = iosgrid_to_wgs84(ng);
    }
    break;
  case BRITAIN:
    {
      BOSGRID ng, g = wgs84_to_bosgrid(pos);
      BOSXY nxy, xy = bosgrid_to_bosxy(g);
      nxy = bosxy_neighbour(xy, e, n);
      ng = bosxy_to_bosgrid(nxy);
      pos = bosgrid_to_wgs84(ng);
    }
  case AEBRITAIN:
    {
      BOSGRID ng, g = wgs84_to_bosgrid(pos);
      BOSXY nxy, xy = bosgrid_to_bosxy(g);
      nxy = bamxy_neighbour(xy, e, n);
      ng = bosxy_to_bosgrid(nxy);
      pos = bosgrid_to_wgs84(ng);
    }
    break;
  }

  new_pos();
}

static int gmv(void)
{
  int *r = (int *) xv_get(motion_speed, PANEL_CLIENT_DATA);
  int i = xv_get(motion_speed, PANEL_VALUE);
  fprintf(stderr, "GMV=%d\n", r[i]);
  return r[i];
}


static void motion_up(void)
{
  do_motion(0, gmv());
}

static void motion_down(void)
{
  do_motion(0, -gmv());
}

static void motion_left(void)
{
  do_motion(-gmv(), 0);
}

static void motion_right(void)
{
  do_motion(gmv(), 0);
}

static void motion_nop(void)
{
  do_motion(0, 0);
}


int heading;
static void canvas_repaint(void)
{
  int w = xv_get(canvas, CANVAS_WIDTH, 0);
  int h = xv_get(canvas, CANVAS_HEIGHT, 0);
  XSetForeground(dpy, gc, 0x000f);
  XFillRectangle(dpy, win, gc, 0, 0, w, h);


  switch (place) {
  case IRELAND:
    i_canvas_repaint();
    break;
  case BRITAIN:
    b_canvas_repaint();
    break;
  case AEBRITAIN:
    bam_canvas_repaint();
    break;
  }

}

static void canvas_null_event(Canvas ec, Event * e)
{
}

static void canvas_ev(Canvas ec, Event * e)
{
  Rect *r;
  int move = 0;

  if (e && event_is_down(e) && (event_action(e) == ACTION_SELECT))
    move++;

  if (e && (!move) && (event_action(e) != LOC_MOVE))
    return;

  r = (Rect *) xv_get(canvas, WIN_MOUSE_XY);

  switch (place) {
  case IRELAND:
    i_clicked(r->r_left, r->r_top, move);
    break;
  case BRITAIN:
    b_clicked(r->r_left, r->r_top, move);
    break;
  case AEBRITAIN:
    bam_clicked(r->r_left, r->r_top, move);
    break;
  }

}

void new_pos(void)
{
  char buf[1024];
  switch (place) {
  case IRELAND:
    i_new_pos();
    break;
  case BRITAIN:
    b_new_pos();
    break;
  case AEBRITAIN:
    bam_new_pos();
    break;
  }
  canvas_ev(canvas, NULL);
  sprintf(buf, "%.8f %.8f", pos.lat, pos.lon);
  xv_set(frame, FRAME_LABEL, buf, 0);
}



float speed_mps = 0;

void update_speeds(float fve, float fvn, float fvu, float a)
{
  char buf[100];
  float h = (180.0 / M_PI) * atan2(fvn, fve);
  int b = (int) (h + .5f);
  float s = sqrt((fvn * fvn) + (fve * fve));

  b = 90 - b;
  while (b < 0)
    b += 360;
  while (b > 360)
    b -= 360;

  heading = b;

  sprintf(buf, "%03d", b);
  xv_set(head, PANEL_VALUE, buf, 0);

  sprintf(buf, "%7.3f", s);
  xv_set(mps, PANEL_VALUE, buf, 0);

  speed_mps *= 0.8;
  speed_mps += 0.2 * s;

  s *= 2.2369363;

  sprintf(buf, "%7.3f", s);
  xv_set(mph, PANEL_VALUE, buf, 0);

  sprintf(buf, "%7.3f", a);
  xv_set(alt, PANEL_VALUE, buf, 0);

}

float ltdiv = 1;

static void place_changed(void)
{
  int i = panel_get_value(placeitem);
  int *j = (int *) xv_get(placeitem, PANEL_CLIENT_DATA);

  float w;

  switch (place) {
  case IRELAND:
    //i_flush_tiles ();
    w = i_get_size();
    break;
  case BRITAIN:
    //b_flush_tiles ();
    w = b_get_size();
    break;
  case AEBRITAIN:
    //bam_flush_tiles ();
    w = bam_get_size();
    break;
  }
  place = (Places) j[i];

  printf("Place is now %d (I=%d,B=%d,A=%d)\n", place, IRELAND, BRITAIN,
         AEBRITAIN);

  switch (place) {
  case IRELAND:
    i_set_size(w);
    break;
  case BRITAIN:
    b_set_size(w);
    break;
  case AEBRITAIN:
    bam_set_size(w);
    break;
  }


}


static void scale_changed(void)
{
  int i = panel_get_value(scale);
  float *j = (float *) xv_get(scale, PANEL_CLIENT_DATA);

  ltdiv = j[i];

  switch (place) {
  case IRELAND:
    i_set_tdiv(j[i]);
    break;
  case BRITAIN:
    b_set_tdiv(j[i]);
    break;
  case AEBRITAIN:
    bam_set_tdiv(j[i]);
    break;
  }

}

Notify_value poke_gps(Notify_client client, int fd)
{
  gps_dispatch();
  return NOTIFY_DONE;
}

int main(int argc, char *argv[])
{
  if ((argc != 2) && (argc != 3) && (argc != 4)) {
    fprintf(stderr, "Usage:\n");
    fprintf(stderr, "map where\n");
    fprintf(stderr, "map where /dev/gpsdevice\n");
    fprintf(stderr, "map where gridreference\n");
    fprintf(stderr, "map where lat lon\n");
    fprintf(stderr, "\n");
    fprintf(stderr,
            "where is B for britain, and I for ireland [including NI]\n");
    fprintf(stderr,
            "british grids are two letters followed by an even \n");
    fprintf(stderr, "	number of digits upto 10\n");
    fprintf(stderr, "irish grids are one letter followed by an even \n");
    fprintf(stderr, "	number of digits upto 10\n");
    fprintf(stderr,
            "once the window comes up you can click to recenter the map\n");
    fprintf(stderr,
            "the gps will do this automatically, lat/lon are in WGS84\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "The environment variables are as follows:\n");
    fprintf(stderr, "BOSPATH path to british OS maps\n");
    fprintf(stderr, "IOSPATH path to irish   OS maps\n");
    fprintf(stderr, "\n");
    exit(1);
  }

  switch (argv[1][0] | 0x20) {
  case 'a':
    place = AEBRITAIN;
    pos.lat = 52.23398100538582554009;
    pos.lon = 0.10222897688425033769;
    break;
  case 'b':
    place = BRITAIN;
    pos.lat = 52.23398100538582554009;
    pos.lon = 0.10222897688425033769;
    break;
  case 'i':
    place = IRELAND;
    pos.lat = 53.2;
    pos.lon = -8.15;
    break;
  default:
    fprintf(stderr,
            "never heard of '%s' where is it, is the weather nice?\n",
            argv[1]);
    exit(1);
  }

  if ((argc == 3) && (argv[2][0] != '/')) {

    if (argv[2][0] == '@') {
      switch (place) {
      case AEBRITAIN:
      case BRITAIN:
        pos = paf_lookup(place, &argv[2][1]);
        break;
      case IRELAND:
        fprintf(stderr, "Postcodes not know for ireland\n");
        break;
      }
    } else {


      switch (place) {
      case AEBRITAIN:
      case BRITAIN:
        {
          BOSGRID g;
          char *ptr = argv[2];
          char e[20], n[20];
          int i, j;

          if (strlen(ptr) % 2) {
            fprintf(stderr, "length of grid ref not even\n");
            exit(1);
          }

          g.big = (*(ptr++)) & ~0x20;
          g.med = (*(ptr++)) & ~0x20;

          bzero(e, sizeof(e));
          bzero(n, sizeof(n));
          strcpy(e, "00000");
          strcpy(n, "00000");

          j = strlen(ptr) >> 1;

          if (j > 5) {
            fprintf(stderr, "too many digits in grid ref\n");
            exit(1);
          }

          for (i = 0; i < j; ++i) {
            e[i] = *(ptr++);
          }
          for (i = 0; i < j; ++i) {
            n[i] = *(ptr++);
          }

          g.e = atoi(e);
          g.n = atoi(n);

          printf("GRID REF: %c%c%s%s\n", g.big, g.med, e, n);

          pos = bosgrid_to_wgs84(g);
        }

        break;
      case IRELAND:
        {
          IOSGRID g;
          char *ptr = argv[2];
          char e[20], n[20];
          int i, j;

          if (!(strlen(ptr) % 2)) {
            fprintf(stderr, "length of grid ref not even\n");
            exit(1);
          }

          g.zone = (*(ptr++)) & ~0x20;

          strcpy(e, "00000");
          strcpy(n, "00000");

          j = strlen(ptr) >> 1;

          if (j > 5) {
            fprintf(stderr, "too many digits in grid ref\n");
            exit(1);
          }

          for (i = 0; i < j; ++i) {
            e[i] = *(ptr++);
          }
          for (i = 0; i < j; ++i) {
            n[i] = *(ptr++);
          }

          g.e = atoi(e);
          g.n = atoi(n);

          printf("GRID REF: %c%s%s\n", g.zone, e, n);

          pos = iosgrid_to_wgs84(g);
        }
        break;
      }
    }
  }

  if (argc == 4) {
    pos.lat = atof(argv[2]);
    pos.lon = atof(argv[3]);
  }


  frame = xv_create((Frame) 0, FRAME, 0);

  dpy = (Display *) xv_get(frame, XV_DISPLAY);
  color_init(frame, dpy);
  gdk_pixbuf_xlib_init_with_depth(dpy, 0, colordepth);


  panel = xv_create(frame, PANEL, XV_HEIGHT, 60, XV_WIDTH, 750, 0);

  canvas = xv_create(frame, CANVAS,
                     WIN_BELOW, panel,
                     XV_X, 0,
                     CANVAS_AUTO_SHRINK, TRUE,
                     CANVAS_AUTO_EXPAND, TRUE,
                     CANVAS_WIDTH, 750, CANVAS_HEIGHT, 550,
                     XV_WIDTH, 750, XV_HEIGHT, 550,
                     CANVAS_X_PAINT_WINDOW, TRUE,
                     CANVAS_REPAINT_PROC, canvas_repaint,
                     WIN_EVENT_PROC, canvas_ev,
                     WIN_CONSUME_EVENTS, LOC_MOVE, LOC_DRAG,
                     WIN_MOUSE_BUTTONS, 0, 0);
  xv_set(canvas_paint_window(canvas),
         WIN_EVENT_PROC, canvas_ev,
         WIN_CONSUME_EVENTS, LOC_MOVE, LOC_DRAG, WIN_MOUSE_BUTTONS, 0, 0);

  head = xv_create(panel, PANEL_TEXT,
                   PANEL_LABEL_STRING, "head:",
                   PANEL_VALUE_DISPLAY_LENGTH, 4, 0);
  mps = xv_create(panel, PANEL_TEXT,
                  PANEL_LABEL_STRING, "m/s:",
                  PANEL_VALUE_DISPLAY_LENGTH, 10, 0);
  mph = xv_create(panel, PANEL_TEXT,
                  PANEL_LABEL_STRING, "m/h:",
                  PANEL_VALUE_DISPLAY_LENGTH, 10, 0);
  alt = xv_create(panel, PANEL_TEXT,
                  PANEL_LABEL_STRING, "Alt:",
                  PANEL_VALUE_DISPLAY_LENGTH, 10, 0);
  {
    static int places[] = { IRELAND, BRITAIN, AEBRITAIN };

    placeitem = xv_create(panel, PANEL_CYCLE,
                          PANEL_LABEL_STRING, "where",
                          PANEL_CHOICE_STRINGS, "Ireland", "Britain(OS)",
                          "Britain(Photo)", NULL, PANEL_CLIENT_DATA,
                          places, PANEL_NOTIFY_PROC, place_changed, 0);
  }
  {
    static float scales[] =
        { 0.166666, 0.2, 0.25, 0.333333, 0.5, 1, 2, 3, 4, 5, 6, 8, 9, 10,
          16, 24, 32, 40,
      48, 56, 64, 72, 80
    };

    scale = xv_create(panel, PANEL_CYCLE,
                      PANEL_LABEL_STRING, "scale",
                      PANEL_CHOICE_STRINGS, "1/6", "1/5", "1/4", "1/3",
                      "1/2", "1", "2", "3", "4", "5", "6", "8", "9", "10",
                      "16", "24", "32", "40", "48", "56", "64", "72", "80",
                      0, PANEL_CLIENT_DATA, scales, PANEL_NOTIFY_PROC,
                      scale_changed, 0);
  }

  gridref = xv_create(panel, PANEL_TEXT,
                      PANEL_LABEL_STRING, "Grid:",
                      PANEL_VALUE_DISPLAY_LENGTH, 18, 0);

  destref = xv_create(panel, PANEL_TEXT,
                      PANEL_LABEL_STRING, "Dest:",
                      PANEL_VALUE_DISPLAY_LENGTH, 18, 0);

  dist = xv_create(panel, PANEL_TEXT,
                   PANEL_LABEL_STRING, "Dist:",
                   PANEL_VALUE_DISPLAY_LENGTH, 10, 0);

  eta = xv_create(panel, PANEL_TEXT,
                  PANEL_LABEL_STRING, "Eta:",
                  PANEL_VALUE_DISPLAY_LENGTH, 10, 0);

  window_fit(frame);

/*
  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "Up", PANEL_NOTIFY_PROC,
             motion_up, 0);
  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "Down",
             PANEL_NOTIFY_PROC, motion_down, 0);
  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "Left",
             PANEL_NOTIFY_PROC, motion_left, 0);
  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "Right",
             PANEL_NOTIFY_PROC, motion_right, 0);
  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, ".", PANEL_NOTIFY_PROC,
             motion_nop, 0);

  {
    static int data[] = { 1, 5, 10, 50, 100, 500, 1000 };

    motion_speed = xv_create (panel, PANEL_CYCLE,
                              PANEL_LABEL_STRING, "speed",
                              PANEL_CHOICE_STRINGS, "1", "5", "10", "50",
                              "100", "500", "1000", NULL, PANEL_CLIENT_DATA,
                              data, 0);
  }
*/

  win = (XID) xv_get(canvas_paint_window(canvas), XV_XID);

  cursor_init();

  {
    XGCValues gc_val;
    gc_val.line_width = 3;
    gc = XCreateGC(dpy, win, GCLineWidth, &gc_val);
  }

  if ((argc == 3) && (argv[2][0] == '/')) {
    int gps = gps_init(argv[2]);

    notify_set_input_func(frame, poke_gps, gps);
  }

  new_pos();
  xv_main_loop(frame);

  return 0;

}
